# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging
import random
from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_features_fid_0(nvme0):
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.getfeatures(0).waitdone()

    
def test_features_sel_00(nvme0):
    orig_config = nvme0.getfeatures(4, sel=0).waitdone()
    nvme0.setfeatures(4, cdw11=orig_config+7).waitdone()
    new_config = nvme0.getfeatures(4, sel=0).waitdone()
    assert new_config == orig_config+7
    nvme0.setfeatures(4, cdw11=orig_config).waitdone()
    

def test_features_sel_01(nvme0, add_value=10):
    if not nvme0.id_data(521, 520)&0x10:
        pytest.skip("feature sv is not supported")

    orig_config = nvme0.getfeatures(4, sel=0).waitdone()
    new_config = nvme0.getfeatures(4, sel=1).waitdone()
    assert orig_config == new_config

    nvme0.setfeatures(4, cdw11=orig_config+add_value).waitdone()
    new_config = nvme0.getfeatures(4, sel=0).waitdone()
    logging.info(orig_config)
    logging.info(new_config)
    assert new_config == orig_config+add_value
    
    new_config = nvme0.getfeatures(4, sel=1).waitdone()
    assert new_config == orig_config

    nvme0.setfeatures(4, cdw11=orig_config).waitdone()


@pytest.mark.xfail(reason="cannot write reserved bit")
def test_features_sel_01_reserved_bit(nvme0):
    test_features_sel_01(nvme0, add_value=10+(1<<1))
    # write a reserved bit
    test_features_sel_01(nvme0, add_value=10+(1<<31))

    
def test_features_sel_10(nvme0, fid=0x10):
    if not nvme0.id_data(521, 520)&0x10:
        pytest.skip("feature sv is not supported")

    # Get random HCTM value
    while(1):
        MNTMT = nvme0.id_data(325, 324)
        MXTMT = nvme0.id_data(327, 326)
        randomTMT1 = random.randrange(MNTMT,MXTMT)
        randomTMT2 = random.randrange(MNTMT,MXTMT)
        if randomTMT1 != randomTMT2:
            break
        else:
            continue
    if randomTMT1 < randomTMT2:
        TMT1 = randomTMT1 << 16
        TMT2 = randomTMT2
    else:
        TMT1 = randomTMT2 << 16
        TMT2 = randomTMT1
    HCTM = TMT1 + TMT2
    
    nvme0.setfeatures(fid, cdw11=HCTM).waitdone()
    
    orig_config = 0
    def getfeatures_cb_1(cdw0, status):
        nonlocal orig_config; orig_config = cdw0
    nvme0.getfeatures(fid, sel=0, cb=getfeatures_cb_1).waitdone()
    logging.info("%x" % orig_config)

    # change value with save bit
    nvme0.setfeatures(fid, sv=1, cdw11=orig_config-1).waitdone()

    new_config = 0
    def getfeatures_cb_2(cdw0, status):
        nonlocal new_config; new_config = cdw0
    nvme0.getfeatures(fid, sel=2, cb=getfeatures_cb_2).waitdone()
    logging.info("%x" % new_config)
    assert new_config == HCTM-1
    nvme0.getfeatures(fid, sel=0, cb=getfeatures_cb_2).waitdone()
    logging.info("%x" % new_config)
    assert new_config == HCTM-1

    # check the feature after reset event
    nvme0.reset()

    new_config = 0
    def getfeatures_cb_3(cdw0, status):
        nonlocal new_config; new_config = cdw0
    nvme0.getfeatures(fid, sel=2, cb=getfeatures_cb_3).waitdone()
    logging.info("%x" % new_config)
    assert new_config == HCTM-1
    nvme0.getfeatures(fid, sel=0, cb=getfeatures_cb_3).waitdone()
    logging.info("%x" % new_config)
    assert new_config == HCTM-1

    # revert to default
    orig_config = 0
    def getfeatures_cb_4(cdw0, status):
        nonlocal orig_config; orig_config = cdw0
    nvme0.getfeatures(fid, sel=1, cb=getfeatures_cb_4).waitdone()
    logging.info("%x" % orig_config)    
    nvme0.setfeatures(fid, cdw11=orig_config).waitdone()
    

def test_features_sel_11(nvme0):
    if not nvme0.id_data(521, 520)&0x10:
        pytest.skip("feature sv is not supported")

    orig_config = 0
    def getfeatures_cb_1(cdw0, status):
        nonlocal orig_config; orig_config = cdw0
    for fid in range(1, 10):
        nvme0.getfeatures(fid, sel=3, cb=getfeatures_cb_1).waitdone()
        logging.info("fid 0x%02x capability: 0x%02x" % (fid, orig_config))


def test_features_invalid_sel(nvme0):
    for f in list(range(1, 0x20)) + list(range(0x80, 0x85)):
        for s in range(4, 8):
            with pytest.warns(UserWarning, match="ERROR status: 00/02"):
                nvme0.getfeatures(fid=f, sel=s).waitdone()

    
def test_features_set_volatile_write_cache(nvme0):
    if not nvme0.id_data(525)&0x1:
        pytest.skip("volatile cache is not preset")

    def get_wce(nvme0):
        wce = 0
        def vwc_cb(cdw0, status):
            nonlocal wce
            wce = cdw0
        nvme0.getfeatures(6, cb=vwc_cb).waitdone()
        return wce

    orig_wce = get_wce(nvme0)
    nvme0.setfeatures(6, cdw11=1).waitdone()
    assert get_wce(nvme0) == 1
    nvme0.setfeatures(6, cdw11=0).waitdone()
    assert get_wce(nvme0) == 0
    nvme0.setfeatures(6, cdw11=orig_wce).waitdone()
    assert orig_wce == get_wce(nvme0)


def test_features_set_invalid_ncqr(nvme0):
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.setfeatures(7, cdw11=0xffff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.setfeatures(7, cdw11=0xffff0000).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.setfeatures(7, cdw11=0xffffffff).waitdone()



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_apst_enabled(nvme0):
    if not nvme0.id_data(265):
        pytest.skip("APST is not enabled")

    pass


def test_host_controlled_thermal_management_enabled(nvme0):
    if not nvme0.id_data(322, 323):
        pytest.skip("APST is not enabled")

    pass


@pytest.mark.parametrize("ps", [4, 3, 2, 1, 0])
def test_format_at_power_state(nvme0, nvme0n1, ps):
    nvme0.setfeatures(0x2, cdw11=ps).waitdone()
    assert nvme0n1.format(ses=0) == 0
    assert nvme0n1.format(ses=1) == 0
    p = nvme0.getfeatures(0x2).waitdone()
    assert p == ps




#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import *


class PRP(Buffer):
    pass


class PRPList(PRP):
    def __init__(self):
        self.prp_per_list = 4096//8
        self.buf_list = [None]*self.prp_per_list

    def __setitem__(self, index, buf: PRP):
        """insert buffer PRP to PRP List"""
        addr = buf.phys_addr
        logging.debug("insert buffer 0x%lx at %d" % (addr, index))
        assert index < self.prp_per_list, "4K PRP List contains 512 PRP entries only"
        self.buf_list[index] = buf

        # fill PRP into PRP List
        for i, b in enumerate(addr.to_bytes(8, byteorder='little')):
            super(PRPList, self).__setitem__(index*8 + i, b)

    def __getitem__(self, index):
        assert index < self.prp_per_list, "4K PRP List contains 512 PRP entries only"
        return self.buf_list[index]

    def find_buffer_by_offset(self, offset, start):
        """find the buffer of the non-contiguous queue contains the offset"""

        first_index = self.offset//8
        for buf in self.buf_list[first_index : self.prp_per_list]:
            assert buf is not None
            if isinstance(buf, PRPList):
                # the last entry could be another PRPList
                assert buf == self.buf_list[self.prp_per_list - 1]
                return buf.find_buffer_by_offset(offset, start)
            else:
                orig_start = start
                start += (len(buf)-buf.offset)
                if start > offset:
                    return buf, buf.offset+offset-orig_start


class SQE(list):
    _buf_list = []

    def __init__(self, *arg):
        assert len(arg) <= 16
        list.extend(self, [0]*16)
        for i, e in enumerate(arg):
            list.__setitem__(self, i, e)

    def __repr__(self):
        return "\n0x%08x, 0x%08x, 0x%08x, 0x%08x" \
               "\n0x%08x, 0x%08x, 0x%08x, 0x%08x" \
               "\n0x%08x, 0x%08x, 0x%08x, 0x%08x" \
               "\n0x%08x, 0x%08x, 0x%08x, 0x%08x" % \
               (self[0], self[1], self[2], self[3], \
                self[4], self[5], self[6], self[7], \
                self[8], self[9], self[10], self[11], \
                self[12], self[13], self[14], self[15])
        
    @property
    def opc(self):
        return self[0]&0xff

    @opc.setter
    def opc(self, opc):
        self[0] = (self[0]&0xffffff00) | opc

    @property
    def cid(self):
        return self[0]>>16

    @cid.setter
    def cid(self, cid):
        self[0] = (self[0]&0xffff) | (cid<<16)

    @property
    def nsid(self):
        return self[1]

    @nsid.setter
    def nsid(self, nsid):
        self[1] = nsid

    @property
    def prp1(self):
        return (self[7]<<32) | self[6]

    @prp1.setter
    def prp1(self, buf: Buffer):
        prp = buf.phys_addr
        self[6] = prp&0xffffffff
        self[7] = (prp>>32)&0xffffffff
        self._buf_list.append(buf)

    @property
    def prp2(self):
        return (self[9]<<32) | self[8]

    @prp2.setter
    def prp2(self, buf: Buffer):
        prp = buf.phys_addr
        self[8] = prp&0xffffffff
        self[9] = (prp>>32)&0xffffffff
        self._buf_list.append(buf)


class CQE(list):
    def __init__(self, arg):
        assert len(arg) == 4
        for e in arg:
            list.append(self, e)

    def __repr__(self):
        return "0x%08x, 0x%08x, 0x%08x, 0x%08x" % \
            (self[0], self[1], self[2], self[3])
        
    @property
    def cdw0(self):
        return self[0]

    @property
    def sqhd(self):
        return self[2]&0xffff

    @property
    def sqid(self):
        return (self[2]>>16)&0xffff

    @property
    def cid(self):
        return self[3]&0xffff

    @property
    def p(self):
        return (self[3]>>16)&0x1

    @property
    def status(self):
        return (self[3]>>17)&0x8fff

    @property
    def sc(self):
        return (self[3]>>17)&0xff

    @property
    def sct(self):
        return (self[3]>>25)&0x7

    @property
    def crd(self):
        return (self[3]>>28)&0x3

    @property
    def m(self):
        return (self[3]>>30)&0x1

    @property
    def dnr(self):
        return (self[3]>>31)&0x1


class IOSQ(object):
    """I/O Submission Queue"""

    id = 0
    ctrlr = None
    queue = None
    sqe_list = None

    def __init__(self, ctrlr, qid, qsize, prp1, pc=True, cqid=None, qprio=0, nvmsetid=0):
        """create IO submission queue

        # Parameters
            ctrlr (Controller):
            qid (int):
            qsize (int): 1based value
            prp1 (PRP, PRPList): the location of the queue
            cqid (int):
            qprio (int):
            nvmsetid (int):

        """

        # some id cq
        status = 0
        if cqid is None:
            cqid = qid

        assert qid < 64*1024 and qid >= 0
        assert cqid < 64*1024 and cqid >= 0
        assert qsize <= 64*1024 and qsize > 0
        assert qprio < 4 and qprio >= 0

        def create_io_sq_cpl(cdw0, status1):
            nonlocal status; status = status1>>1

        self.ctrlr = ctrlr
        ctrlr.send_cmd(0x01, prp1,
                       cdw10 = (qid|((qsize-1)<<16)),
                       cdw11 = (pc|(qprio<<1)|(cqid<<16)),
                       cdw12 = nvmsetid,
                       cb = create_io_sq_cpl).waitdone()
        if not status:
            self.id = qid
            self.queue = prp1
            self.sqe_list = [None]*qsize

    def __getitem__(self, index):
        assert index < len(self.sqe_list)
        return self.sqe_list[index]
    
    def __setitem__(self, index, cmd: SQE):
        """insert 16-dword SQE to the queue"""

        assert len(cmd) == 16
        assert index < len(self.sqe_list)

        # track the cmd in the queue
        self.sqe_list[index] = cmd

        # locate the queue buffer
        if isinstance(self.queue, PRPList):
            # find the PRP entry in non-contig queue
            buf, offset = self.queue.find_buffer_by_offset(index*64, 0)
            index = offset/64
        else:
            buf = self.queue

        assert isinstance(buf, PRP)
        for i, dword in enumerate(cmd):
            for j, b in enumerate(dword.to_bytes(4, byteorder='little')):
                buf[index*64 + i*4 + j] = b

    @property
    def tail(self):
        return self.ctrlr[0x1000+2*self.id*4]

    @tail.setter
    def tail(self, tail):
        logging.debug(tail)
        self.ctrlr[0x1000+2*self.id*4] = tail

    def delete(self, qid=None):
        if qid == None:
            qid = self.id

        logging.debug("delete sqid %d" % qid)
        if qid != 0:
            self.ctrlr.send_cmd(0x00, cdw10=qid).waitdone()


class IOCQ(object):
    """I/O Completion Queue"""

    id = 0
    ctrlr = None
    queue = None
    qsize = 0

    def __init__(self, ctrlr, qid, qsize, prp1, pc=True, iv=0, ien=False):
        """create IO completion queue

        # Parameters
            ctrlr (Controller):
            qid (int):
            qsize (int): 1based value
            prp1 (Buffer, PRPList):
            iv (int, None): interrupt vector

        """

        status = 0

        assert qid < 64*1024 and qid >= 0
        assert qsize < 64*1024 and qsize > 0

        def create_io_cq_cpl(cdw0, status1):
            nonlocal status; status = status1>>1

        self.ctrlr = ctrlr
        ctrlr.send_cmd(0x05, prp1,
                       cdw10 = (qid|((qsize-1)<<16)),
                       cdw11 = (pc|(ien<<1)|(iv<<16)),
                       cb = create_io_cq_cpl).waitdone()

        if not status:
            self.id = qid
            self.queue = prp1
            self.qsize = qsize

    def __getitem__(self, index):
        """get 4-dword CQE from the queue"""

        assert index < self.qsize

        cpl = [0]*4
        buf = self.queue
        if isinstance(buf, PRPList):
            # find the PRP entry in non-contig queue
            buf, offset = self.queue.find_buffer_by_offset(index*16, 0)
            index = offset/16
        else:
            buf = self.queue

        assert isinstance(buf, PRP)
        for i in range(4):
            cpl[i] = buf.data(index*16 + i*4 + 3, index*16 + i*4)

        assert len(cpl) == 4
        return CQE(cpl)

    @property
    def head(self):
        return self.ctrlr[0x1000+(2*self.id+1)*4]

    @head.setter
    def head(self, head):
        self.ctrlr[0x1000+(2*self.id+1)*4] = head

    def delete(self, qid=None):
        if qid is None:
            qid = self.id

        logging.debug("delete cqid %d" % qid)
        if qid != 0:
            self.ctrlr.send_cmd(0x04, cdw10=qid).waitdone()


def test_create_delete_iocq(nvme0):
    buf = Buffer(4096)

    cq1 = IOCQ(nvme0, 5, 5, buf)

    # Invalid Queue Identifier
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        cq2 = IOCQ(nvme0, 5, 10, buf)

    cq1.delete()

    # Invalid Queue Size
    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        cq = IOCQ(nvme0, 5, 1, buf)

    # Invalid Queue Identifier
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        cq = IOCQ(nvme0, 0, 5, buf)

    # Invalid interrupt vector
    with pytest.warns(UserWarning, match="ERROR status: 01/08"):
        cq = IOCQ(nvme0, 5, 5, buf, iv=2047, ien=True)

    # Invalid interrupt vector
    with pytest.warns(UserWarning, match="ERROR status: 01/08"):
        cq = IOCQ(nvme0, 5, 5, buf, iv=2049, ien=True)

    cq = IOCQ(nvme0, 5, 5, buf, iv=5)
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        cq.delete(4)
    cq.delete()

    cq = IOCQ(nvme0, 5, 5, buf, iv=5, ien=True)
    cq.delete()

    cq1 = IOCQ(nvme0, 5, 5, buf)
    cq2 = IOCQ(nvme0, 6, 5, buf)
    cq3 = IOCQ(nvme0, 7, 5, buf)
    cq3.delete()
    cq2.delete()
    cq1.delete()


@pytest.mark.parametrize("pgsz", [1, 2, 3, 10, 256, 512, 1024])
def test_create_delete_iocq_large(nvme0, pgsz):
    buf_cq = PRP(4096*pgsz)
    cq = IOCQ(nvme0, 4, 5, buf_cq)
    cq.delete()


def test_create_delete_iocq_non_contig(nvme0):
    prp_list = PRPList()
    prp_list[0] = PRP()
    prp_list[1] = PRP()

    cq = IOCQ(nvme0, 4, 5, prp_list, pc=False)
    cq.delete()


def test_create_delete_iosq(nvme0):
    buf_cq = PRP(4096)
    cq = IOCQ(nvme0, 4, 5, buf_cq)

    buf_sq = PRP(4096)

    # Completion Queue Invalid
    with pytest.warns(UserWarning, match="ERROR status: 01/00"):
        sq = IOSQ(nvme0, 4, 10, buf_sq, cqid=1)

    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        sq = IOSQ(nvme0, 400, 10, buf_sq, cqid=4)

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        sq = IOSQ(nvme0, 4, 1, buf_sq, cqid=4)

    sq = IOSQ(nvme0, 5, 10, buf_sq, cqid=4)
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        sq.delete(4)

    # Invalid Queue Deletion
    with pytest.warns(UserWarning, match="ERROR status: 01/0c"):
        cq.delete()

    sq.delete()
    cq.delete()


def test_send_single_cmd(nvme0):
    cq = IOCQ(nvme0, 1, 10, PRP())
    sq = IOSQ(nvme0, 1, 10, PRP(), cqid=1)

    # first cmd, invalid namespace
    sq[0] = [4] + [0]*15
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x000b

    sq.delete()
    cq.delete()


@pytest.mark.parametrize("count", [1, 2, 8, 500, 512])
def test_prp_and_prp_list(count):
    l = PRPList()
    for i in range(count):
        l[i] = Buffer()


def test_prp_and_prp_list_with_offset():
    p = PRP()
    p.offset = 0x20

    l = PRPList()
    l.offset = 0x40
    l[8] = p

    p = PRP()
    p.offset = 0x30
    l[9] = p

    assert l.phys_addr&0x7 == 0
    assert l.phys_addr&0xfff == 0x40
    assert l[8].phys_addr&0xfff == 0x20
    assert l[9].phys_addr&0xfff == 0x30


def test_prp_and_prp_list_invalid():
    l = PRPList()
    with pytest.raises(AssertionError):
        l[512] = Buffer()

        
@pytest.mark.parametrize("qdepth", [7, 2, 3, 4, 5, 10, 16, 17, 31])
def test_send_cmd_different_qdepth(nvme0, qdepth):
    cq = IOCQ(nvme0, 4, qdepth, PRP())
    sq = IOSQ(nvme0, 4, qdepth, PRP(), cqid=4)

    # once again: first cmd, invalid namespace
    for i in range(qdepth*3 + 3):
        index = (i+1)%qdepth
        sq[index-1] = [0, 1] + [0]*14
        sq.tail = index
        time.sleep(0.01)
        assert (cq[index-1][3]>>17) == 0
        assert (cq[index-1][2]&0xffff) == index
        cq.head = index

    sq.delete()
    cq.delete()
    

def test_send_cmd_2sq_1cq(nvme0):
    cq = IOCQ(nvme0, 1, 10, PRP())
    sq1 = IOSQ(nvme0, 1, 10, PRP(), cqid=1)
    sq2 = IOSQ(nvme0, 2, 16, PRP(), cqid=1)

    cdw = SQE(0, 0, 0)
    cdw.nsid = 1  # namespace id
    cdw.cid = 222
    sq1[0] = cdw

    sqe = SQE(*cdw)
    assert sqe[1] == 1
    sqe.cid = 111
    sq2[0] = sqe
    sq2.tail = 1
    time.sleep(0.1)
    sq1.tail = 1
    time.sleep(0.1)

    sq1.delete()
    sq2.delete()
    cq.delete()

        
def test_psd_write_2sq_1cq_prp_list(nvme0, subsystem):
    # cqid: 1, PC, depth: 120
    cq = IOCQ(nvme0, 1, 120, PRP(4096))

    # create two SQ, both use the same CQ
    # sqid: 3, depth: 16
    sq3 = IOSQ(nvme0, 3, 16, PRP(), cqid=1)
    # sqid: 5, depth: 100, so need 2 pages of memory
    sq5 = IOSQ(nvme0, 5, 64*64, PRP(4096*64), cqid=1)

    # IO command templates: opcode and namespace
    write_cmd = SQE(1, 1)
    read_cmd = SQE(2, 1)

    # write in sq3, lba1-lba2, 1 page, aligned
    w1 = SQE(*write_cmd)
    write_buf = PRP(ptype=32, pvalue=0xaaaaaaaa)
    w1.prp1 = write_buf
    w1[10] = 1
    w1[12] = 1 # 0based
    w1.cid = 0x123
    sq3[0] = w1
    sq3.tail = 1

    # add some delay, so ssd should finish w1 before w2
    time.sleep(0.1)

    # write in sq5, lba5-lba16, 2 page, non aligned
    w2 = SQE(*write_cmd)
    buf1 = PRP(ptype=32, pvalue=0xbbbbbbbd)
    buf1.offset = 2048
    w2.prp1 = buf1
    w2.prp2 = PRP(ptype=32, pvalue=0xcccccccc)
    w2[10] = 5
    w2[12] = 11 # 0based
    w2.cid = 0x567
    sq5[0] = w2
    sq5.tail = 1

    # cqe for w1
    while CQE(cq[0]).p == 0: pass
    cqe = CQE(cq[0])
    assert cqe.cid == 0x123
    assert cqe.sqid == 3
    assert cqe.sqhd == 1
    cq.head = 1

    # cqe for w2
    while CQE(cq[1]).p == 0: pass
    cqe = CQE(cq[1])
    assert cqe.cid == 0x567
    assert cqe.sqid == 5
    assert cqe.sqhd == 1
    cq.head = 2

    # read in sq3, lba0-lba23, 3 page with PRP list
    r1 = SQE(*read_cmd)
    read_buf = [PRP() for i in range(3)]
    r1.prp1 = read_buf[0]
    prp_list = PRPList()
    prp_list[0] = read_buf[1]
    prp_list[1] = read_buf[2]
    r1.prp2 = prp_list
    r1[10] = 0
    r1[12] = 23 # 0based
    sq3[1] = r1
    sq3.tail = 2

    # verify read data
    while cq[2].p == 0: pass
    cq.head = 3
    assert read_buf[0].data(0xfff, 0xffc) == 0xbbbbbbbd
    assert read_buf[2].data(3, 0) == 0xcccccccc
    assert read_buf[2].data(0x1ff, 0x1fc) == 0xcccccccc

    # delete all sq/cq
    sq3.delete()
    sq5.delete()
    cq.delete()


def test_iocq_prplist():
    prp_list = PRPList()
    p = PRP()
    p.offset = 4096-64
    prp_list[510] = p
    prp_list.offset = 4096-16

    prp_list2 = PRPList()
    prp_list[511] = prp_list2

    prp_list2[0] = PRP()
    prp_list2[1] = PRP()
    prp_list2[2] = PRP()
    prp_list2[3] = PRP()

    buffer, offset = prp_list.find_buffer_by_offset(0, 0)
    assert buffer == p
    assert offset == p.offset

    buffer, offset = prp_list.find_buffer_by_offset(5*16, 0)
    assert buffer == prp_list2[0]
    assert offset == 16

    buffer, offset = prp_list.find_buffer_by_offset(1000*16, 0)
    assert buffer == prp_list2[3]
    assert offset == 3648


def test_psd_with_qpair(nvme0):
    # do not mix psd IO queues with SPDK qpairs in the same test script
    qpair = Qpair(nvme0, 16)
    buf_cq = PRP(4096)
    qid = qpair.sqid
    # Invalid Queue Identifier: the id is occupied
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        cq = IOCQ(nvme0, qid, 5, buf_cq)
    cq.delete()
    qpair.delete()

    cq = IOCQ(nvme0, qid, 5, buf_cq)
    # the qid 1 was occupied by psd IOCQ first
    with pytest.raises(QpairCreationError):
        qpair = Qpair(nvme0, 16)
    cq.delete()


def test_write_before_power_cycle(nvme0, subsystem):
    cq = IOCQ(nvme0, 1, 128, PRP(2*1024))
    sq = IOSQ(nvme0, 1, 128, PRP(8*1024), cqid=1)

    #burst write
    for i in range(127):
        cmd = SQE(1, 1)
        buf = PRP(512, ptype=32, pvalue=i)
        cmd.prp1 = buf
        cmd[10] = i
        sq[i] = cmd

    # write 127 512byte at one shot
    sq.tail = 127

    #not to wait commands completion
    # while CQE(cq[126]).p == 0: pass
    # cq.head = 0

    # sq.delete()
    # cq.delete()

    # power off immediately without completion of the sub process
    subsystem.power_cycle(10)
    nvme0.reset()

    # read and check
    cq = IOCQ(nvme0, 2, 16, PRP())
    sq = IOSQ(nvme0, 2, 16, PRP(), cqid=2)

    cmd = SQE(2, 1)
    buf_read = PRP()
    cmd.prp1 = buf_read
    cmd[10] = 126
    sq[0] = cmd
    sq.tail = 1
    while CQE(cq[0]).p == 0: pass
    cq.head = 1
    print(buf_read.dump(32))

    sq.delete()
    cq.delete()


def test_invalid_sq_doorbell(nvme0):
    cq = IOCQ(nvme0, 4, 16, PRP())
    sq1 = IOSQ(nvme0, 4, 16, PRP(), cqid=4)

    write_cmd = SQE(1, 1)
    write_cmd.prp1 = PRP()
    prp_list = PRPList()
    prp_list[0] = PRP()
    prp_list[1] = PRP()
    prp_list[2] = PRP()
    write_cmd.prp2 = prp_list
    write_cmd[10] = 0
    write_cmd[12] = 31
    write_cmd.cid = 123

    sq1[0] = write_cmd
    write_cmd.cid = 567
    sq1.tail = 17

    # wait for the controller to respond the error
    time.sleep(0.1)
    with pytest.warns(UserWarning, match="AER notification is triggered: 0x10100"):
        nvme0.getfeatures(7).waitdone()
    sq1.delete()
    cq.delete()
    

def test_write_read_verify(nvme0, nvme0n1):
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    sq = IOSQ(nvme0, 1, 1024, PRP(1024*64), cqid=1)

    write_cmd = SQE(1, 1)
    write_cmd[12] = 7  # 4K write
    buf_list = []
    for i in range(10):
        buf = PRP(ptype=32, pvalue=0x01010101*i)
        buf_list.append(buf)
        write_cmd.prp1 = buf
        write_cmd.cid = i
        write_cmd[10] = i
        sq[i] = write_cmd
    sq.tail = 10
    
    while cq[9].p == 0: pass
    sq.delete()
    cq.delete()
    
    # read after reset with outstanding writes
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    sq = IOSQ(nvme0, 1, 1024, PRP(1024*64), cqid=1)

    read_cmd = SQE(2, 1)
    read_cmd[12] = 7  # 4K read
    buf_list = []
    for i in range(10):
        buf = PRP()
        buf_list.append(buf)
        read_cmd.prp1 = buf
        read_cmd.cid = i
        read_cmd[10] = i
        sq[i] = read_cmd
    sq.tail = 10

    while cq[9].p == 0: pass
    for i in range(10):
        logging.info(cq[i].cid)
        assert buf_list[cq[i].cid][0] == cq[i].cid
    sq.delete()
    cq.delete()





# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_verify_valid(nvme0, nvme0n1):
    if not nvme0n1.supports(0xc):
        pytest.skip("verify command is not supported")

    pass





# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def nvme_init_wrr(nvme0):
    logging.info("user defined nvme init")

    nvme0[0x14] = 0
    while not (nvme0[0x1c]&0x1) == 0: pass

    # 3. set admin queue registers
    nvme0.init_adminq()

    # 4. set register cc
    if (nvme0.cap>>17) & 0x1:
        logging.info("set WRR arbitration")
        nvme0[0x14] = 0x00460800
    else:
        nvme0[0x14] = 0x00460000

    # 5. enable cc.en
    nvme0[0x14] = nvme0[0x14] | 1

    # 6. wait csts.rdy to 1
    while not (nvme0[0x1c]&0x1) == 1: pass

    # 7. identify controller
    nvme0.identify(Buffer(4096)).waitdone()

    # 8. create and identify all namespace
    nvme0.init_ns()

    # 9. set/get num of queues
    logging.debug("init number of queues")
    nvme0.setfeatures(0x7, cdw11=0xfffefffe).waitdone()
    cdw0 = nvme0.getfeatures(0x7).waitdone()
    nvme0.init_queues(cdw0)

@pytest.fixture()
def nvme0(pcie):
    return Controller(pcie, nvme_init_func=nvme_init_wrr)

@pytest.fixture()
def nvme0n1(nvme0):
    ret = Namespace(nvme0, 1, 0x10000)
    yield ret
    ret.close()


def test_ioworker_with_wrr(nvme0, nvme0n1):
    if (nvme0.cap>>17) & 0x1 == 0:
        pytest.skip("WRR is not supported")

    nvme0n1.format(512)

    # 8:4:2
    assert nvme0[0x14] == 0x00460801
    nvme0.setfeatures(1, cdw11=0x07030103).waitdone()
    cdw0 = nvme0.getfeatures(1).waitdone()
    assert cdw0 == 0x07030103

    l = []
    for i in range(3):
        a = nvme0n1.ioworker(io_size=8,
                             read_percentage=100,
                             region_end=0x10000,
                             qprio=i+1,
                             time=30)
        l.append(a)

    w = []
    for a in l:
        r = a.start()
        w.append(r)

    io_count = []
    for a in l:
        r = a.close()
        logging.debug(r)
        io_count.append(r.io_count_read)

    logging.info(io_count)
    assert io_count[1]/io_count[2] > 1.7
    assert io_count[1]/io_count[2] < 2.2
    assert io_count[0]/io_count[1] > 1.7
    assert io_count[0]/io_count[1] < 2.2


def test_weighed_round_robin(nvme0):
    if (nvme0.cap>>17) & 0x1 == 0:
        pytest.skip("WRR is not supported")

    assert nvme0[0x14] == 0x00460801
    nvme0.setfeatures(1, cdw11=0x07030103).waitdone()
    cdw0 = nvme0.getfeatures(1).waitdone()
    assert cdw0 == 0x07030103

    start = time.time()
    nvme0.getfeatures(7).waitdone()
    logging.info("admin latency: %f" % (time.time()-start))

    # create the senario of Figure 138 in NVMe spec v1.4
    # 1 admin, 2 urgent, 2 high, 2 medium, 2 low
    sq_list = []
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    for i in range(8):
        sq_list.append(IOSQ(nvme0, i+1, 128, PRP(128*64), cqid=1, qprio=i//2))

    # fill 100 flush commands in each queue
    for sq in sq_list:
        for i in range(100):
            sq[i] = SQE(i<<16+0, 1)

    # fire all sq, low prio first
    for sq in sq_list[::-1]:
        sq.tail = 100

    # check the latency of admin command
    start = time.time()
    nvme0.getfeatures(7).waitdone()
    logging.info("admin latency when IO busy: %f" % (time.time()-start))

    # check sqid of the whole cq
    time.sleep(1)
    logging.info([cq[i][2]>>16 for i in range(100*8)])
    # assert all urgent IO completed first
    last_sqid = {cq[i][2]>>16 for i in range(300, 800)}
    assert 1 not in last_sqid
    assert 2 not in last_sqid

    # delete all queues
    for sq in sq_list:
        sq.delete()
    cq.delete()


def test_default_round_robin(nvme0):
    # 1 admin, 8 io queue
    sq_list = []
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    for i in range(8):
        sq_list.append(IOSQ(nvme0, i+1, 128, PRP(128*64), cqid=1))

    # fill 100 flush commands in each queue
    for sq in sq_list:
        for i in range(100):
            sq[i] = SQE(i<<16+0, 1)

    # fire all sq, low prio first
    for sq in sq_list:
        sq.tail = 100

    # check the latency of admin command
    start = time.time()
    nvme0.getfeatures(7).waitdone()
    logging.info("admin latency when IO busy: %f" % (time.time()-start))

    # check sqid of the whole cq
    time.sleep(1)
    logging.debug([cq[i][2]>>16 for i in range(100*8)])
    # assert all urgent IO completed first
    last_sqid = {cq[i][2]>>16 for i in range(700, 800)}
    assert len(last_sqid) >= 7

    # delete all queues
    for sq in sq_list:
        sq.delete()
    cq.delete()



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import warnings

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem, __version__


def test_dut_firmware_and_model_name(nvme0):
    logging.info(nvme0.id_data(63, 24, str))
    logging.info(nvme0.id_data(71, 64, str))
    logging.info("testing conformance with pynvme " + __version__)

    
def test_abort_all_aer_commands(nvme0):
    aerl = nvme0.id_data(259)+1
    logging.info(aerl)

    def aer_cb(cdw0, status):
        assert ((status&0xfff)>>1) == 0x0007
    # another one is sent in defaul nvme init
    for i in range(aerl-1):
        nvme0.aer(cb=aer_cb)

    # ASYNC LIMIT EXCEEDED (01/05)
    with pytest.warns(UserWarning, match="ERROR status: 01/05"):
        nvme0.aer()
        nvme0.getfeatures(7).waitdone()

    # no timeout happen on aer
    time.sleep(15)

    # ABORTED - BY REQUEST (00/07)
    logging.info("reap %d command, including abort, and also aer commands" % (100+aerl))
    with pytest.warns(UserWarning, match="ERROR status: 00/07"):
        for i in range(100):
            nvme0.abort(127-i).waitdone()
        

def test_abort_specific_aer_command(nvme0):
    aborted = False
    def aer_cb(cpl):
        nonlocal aborted
        status = cpl[3]>>16
        if ((status&0xfff)>>1) == 0x0007:
            aborted = True
    nvme0.aer(cb=aer_cb)
    with pytest.warns(UserWarning, match="ERROR status: 00/07"):
        nvme0.abort(nvme0.latest_cid).waitdone()
    assert aborted


def test_abort_abort_command(nvme0):
    nvme0.abort(0)
    nvme0.abort(nvme0.latest_cid)
    nvme0.waitdone(2)

    nvme0.abort(0xffff)
    nvme0.abort(nvme0.latest_cid)
    nvme0.waitdone(2)
    
    nvme0.aer()
    nvme0.abort(nvme0.latest_cid) # abort aer command
    nvme0.abort(nvme0.latest_cid) # abort abort command
    with pytest.warns(UserWarning, match="ERROR status: 00/07"):
        nvme0.waitdone(2)
    



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_getlogpage_page_id(nvme0, buf):
    for lid in (1, 2, 3):
        nvme0.getlogpage(lid, buf).waitdone()

    for lid in (0, 0x6f, 0x7f):
        with pytest.warns(UserWarning, match="ERROR status: 01/09"):
            nvme0.getlogpage(lid, buf).waitdone()
        

@pytest.mark.parametrize("repeat", range(32))
def test_getlogpage_invalid_numd(nvme0, repeat):
    dts = nvme0.mdts//4 + 1 + repeat
    buf = Buffer(dts*4)

    for lid in (1, 2, 3):
        nvme0.getlogpage(lid, buf).waitdone()


def test_getlogpage_after_error(nvme0, nvme0n1, buf, qpair):
    nvme0.getlogpage(1, buf).waitdone()
    assert buf.data(7, 0) == 0
    nvme0n1.write_uncorrectable(qpair, 0, 8).waitdone()

    # generate 2 errors
    with pytest.warns(UserWarning, match="ERROR status: 02/81"):
        nvme0n1.read(qpair, buf, 0, 8).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/81"):
        nvme0n1.read(qpair, buf, 0, 8).waitdone()

    time.sleep(0.1) # wait error information ready
    nvme0.getlogpage(1, buf).waitdone()
    nerror1 = buf.data(7, 0)
    nerror2 = buf.data(64+7, 64)
    assert nerror1 == nerror2+1

    nvme0n1.write(qpair, buf, 0, 8).waitdone()
    

@pytest.mark.parametrize("len", (1, 2, 4))
def test_getlogpage_data_unit_read(nvme0, nvme0n1, buf, qpair, len):
    if not nvme0n1.supports(5):
        pytest.skip("compare is not support")

    nvme0n1.write(qpair, buf, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nread1 = buf.data(47, 32)

    # read
    for i in range(1000):
        nvme0n1.read(qpair, buf, 0, len).waitdone()
    nvme0.getlogpage(2, buf).waitdone()
    nread2 = buf.data(47, 32)
    assert nread2 == nread1+len

    # compare
    nvme0n1.read(qpair, buf, 0, len).waitdone()  # get correct data
    for i in range(1000):
        nvme0n1.compare(qpair, buf, 0, len).waitdone()
    nvme0.getlogpage(2, buf).waitdone()
    nread3 = buf.data(47, 32)
    assert nread3 == nread2+len

    # verify
    if nvme0n1.supports(0xc):
        for i in range(1000):
            nvme0n1.verify(qpair, 0, len).waitdone()
        nvme0.getlogpage(2, buf).waitdone()
        nread4 = buf.data(47, 32)
        assert nread4 == nread3+len    


@pytest.mark.parametrize("len", (1, 2, 4))
def test_getlogpage_data_unit_write(nvme0, nvme0n1, len, buf, qpair):
    if not nvme0n1.supports(5):
        pytest.skip("compare is not support")

    nvme0n1.write(qpair, buf, 0).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite1 = buf.data(63, 48)

    for i in range(1000):
        nvme0n1.write(qpair, buf, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite2 = buf.data(63, 48)
    assert nwrite2 == nwrite1+len
    
    for i in range(1000):
        nvme0n1.write_uncorrectable(qpair, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite3 = buf.data(63, 48)
    assert nwrite3 == nwrite2

    nvme0n1.write(qpair, Buffer(4096), 0, 8).waitdone()
    

def test_getlogpage_power_cycle_count(nvme0, subsystem, buf):
    def get_power_cycles(nvme0):
        nvme0.getlogpage(2, buf, 512).waitdone()
        return buf.data(115, 112)

    powercycle = get_power_cycles(nvme0)
    subsystem.power_cycle(10)
    nvme0.reset()
    assert get_power_cycles(nvme0) == powercycle+1


def test_getlogpage_namespace(nvme0, buf):
    nvme0.getlogpage(2, buf, nsid=1).waitdone()
    nvme0.getlogpage(2, buf, nsid=0xffffffff).waitdone()

    # getlogpage for id=2 nsid can be 0 or 0xffffffff
    nvme0.getlogpage(2, buf, nsid=0).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.getlogpage(2, buf, nsid=2).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.getlogpage(2, buf, nsid=0xfffffffe).waitdone()

        
def test_getlogpage_smart_composite_temperature(nvme0):
    smart_log = Buffer()
    
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.debug("temperature: %d degreeF" % ktemp)

    # warning with AER
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        # over composite temperature threshold
        nvme0.setfeatures(4, cdw11=ktemp-10).waitdone()
        
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()
        logging.debug("0x%x" % smart_log.data(0))
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()    
        assert smart_log.data(0) & 0x2

        # higher threshold
        nvme0.setfeatures(4, cdw11=ktemp+10).waitdone()

    # aer is not expected
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.debug("temperature: %d degreeF" % ktemp)

    # revert to default
    orig_config = 0
    def getfeatures_cb_4(cdw0, status):
        nonlocal orig_config; orig_config = cdw0
    nvme0.getfeatures(4, sel=1, cb=getfeatures_cb_4).waitdone()
    nvme0.setfeatures(4, cdw11=orig_config).waitdone()    

    
def test_getlogpage_persistent_event_log(nvme0):
    if not (nvme0.id_data(261)&0x10):
        pytest.skip("feature sv is not supported")


def test_getlogpage_firmware_slot_info_nsid_1(nvme0, buf):
    """For Log Pages with a scope of NVM subsystem or controller (as shown in Figure 191 and Figure 192), the
controller should abort commands that specify namespace identifiers other than 0h or FFFFFFFFh with
status Invalid Field in Command."""

    nvme0.getlogpage(3, buf, 512, nsid=0).waitdone()
    nvme0.getlogpage(3, buf, 512, nsid=0xffffffff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.getlogpage(3, buf, 512, nsid=1).waitdone()



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


@pytest.mark.parametrize("size", [4096, 10, 4096*2])
@pytest.mark.parametrize("offset", [4096, 10, 4096*2])
def test_firmware_download(nvme0, size, offset):
    fwug = nvme0.id_data(319)
    logging.info(fwug)
    if fwug == 0 or fwug == 0xff:
        pytest.skip("not applicable")
    
    buf = Buffer(size*fwug)
    with pytest.warns(UserWarning, match="ERROR status: (01/14|00/02)"):
        nvme0.fw_download(buf, offset*fwug).waitdone()


def test_firmware_commit(nvme0):
    frmw = nvme0.id_data(260)
    slot1_ro = frmw&1
    slot_count = (frmw>>1)&7
    logging.info(slot1_ro)
    logging.info(slot_count)
    
    buf = Buffer(4096*16)
    nvme0.fw_download(buf, 0).waitdone()

    logging.info("commit to invalid firmware slot")
    with pytest.warns(UserWarning, match="ERROR status: 01/07"):
        nvme0.fw_commit(0, 0).waitdone()

    if slot_count != 7:
        for slot in range(slot_count+1, 8):
            with pytest.warns(UserWarning, match="ERROR status: 01/06"):
                nvme0.fw_commit(slot, 2).waitdone()

    logging.info("commit with an invalid firmware image")
    for slot in range(2, slot_count+1):
        with pytest.warns(UserWarning, match="ERROR status: 01/07"):
            nvme0.fw_commit(slot, 0).waitdone()
        
    with pytest.warns(UserWarning, match="ERROR status: 01/07"):
        nvme0.fw_commit(1, 0).waitdone()
